"""
Neural Additive Bayesian Quantile Regression (NABQR) Implementation

This module provides the main interface for running NABQR models. It combines
ensemble generation, model training, and visualization capabilities.
"""

from functions import *
from helper_functions import simulate_correlated_ar1_process, set_n_closest_to_zero
import matplotlib.pyplot as plt
import scienceplots
plt.style.use(['no-latex'])
from visualization import visualize_results 
import datetime as dt

def run_nabqr_pipeline(n_samples=5000, phi=0.995, sigma=8, offset_start=10, offset_end=500, 
                      offset_step=15, correlation=0.8, data_source="NABQR-TEST", 
                      training_size=0.7, epochs=100, timesteps=[0,1,2,6,12,24],
                      quantiles=[0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 0.99]):
    """
    Runs the complete NABQR pipeline including data simulation, model training, and visualization.
    
    Parameters:
    n_samples (int): Number of time steps to simulate
    phi (float): AR(1) coefficient for simulation
    sigma (float): Standard deviation of noise for simulation
    offset_start (int): Start value for offset range
    offset_end (int): End value for offset range
    offset_step (int): Step size for offset range
    correlation (float): Base correlation between dimensions
    data_source (str): Identifier for the data source
    training_size (float): Proportion of data to use for training
    epochs (int): Number of epochs for model training
    timesteps (list): List of timesteps to use for LSTM
    quantiles (list): List of quantiles to predict
    
    Returns:
    None: Saves results to files and displays visualization
    """
    # Generate offset and correlation matrix
    offset = np.arange(offset_start, offset_end, offset_step)
    m = len(offset)
    corr_matrix = correlation * np.ones((m, m)) + (1-correlation) * np.eye(m)
    
    # Generate simulated data
    simulated_data, actuals = simulate_correlated_ar1_process(n_samples, phi, sigma, 
                                                            m, corr_matrix, offset, smooth=5)
    
    # Run the pipeline
    pipeline(simulated_data, actuals, data_source, training_size=training_size, 
            epochs=epochs, timesteps_for_lstm=timesteps, quantiles_taqr=quantiles)
    
    # Get today's date for file naming
    today = dt.datetime.today().strftime('%Y-%m-%d')
    
    # Load results
    CE = pd.read_csv(f"results_{today}_{data_source}_corrected_ensembles.csv")
    y_hat = np.load(f"results_{today}_{data_source}_actuals_out_of_sample.npy")
    q_hat = np.load(f"results_{today}_{data_source}_taqr_results.npy")
    
    # Visualize results
    visualize_results(y_hat, q_hat, f"{data_source} example")


"""
Neural Additive Bayesian Quantile Regression (NABQR) Core Functions

This module provides the core functionality for NABQR, including:
- Scoring metrics (Variogram, CRPS, QSS)
- Dataset creation and preprocessing
- Model definitions and training
- TAQR (Time-Adaptive Quantile Regression) implementation
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import properscoring as ps
import tensorflow as tf
import tensorflow_probability as tfp
import datetime as dt
from helper_functions import simulate_correlated_ar1_process, set_n_closest_to_zero
from functions_for_TAQR import *

def variogram_score_single_observation(x, y, p=0.5):
    """
    Calculate the Variogram score for a given observation.
    Translated from the R code in Energy and AI paper.

    Parameters:
    x (numpy.ndarray): Ensemble forecast (m x k), where m is ensemble size, k is forecast horizon
    y (numpy.ndarray): Actual observations (k,)
    p (float): Power parameter for the variogram score

    Returns:
    float: Variogram score for the observation
    """
    m, k = x.shape
    score = 0

    for i in range(k - 1):
        for j in range(i + 1, k):
            Ediff = (1 / m) * np.sum(np.abs(x[:, i] - x[:, j])**p)
            score += (1/np.abs(i-j))*(np.abs(y[i] - y[j])**p - Ediff)**2

    return score / k

def variogram_score_R_multivariate(x, y, p=0.5, t1=12, t2=36):
    """
    Calculate the Variogram score for all observations for the time horizon t1 to t2.
    
    Parameters:
    x (numpy.ndarray): Ensemble forecast (m x k)
    y (numpy.ndarray): Actual observations (k,)
    p (float): Power parameter
    t1 (int): Start hour (inclusive)
    t2 (int): End hour (exclusive)
    
    Returns:
    tuple: (score, score_list) Overall score and list of individual scores
    """
    m, k = x.shape
    score = 0
    if m > k:
        x = x.T
        m, k = k, m
    
    score_list = []
    for start in range(0, k, 24):
        if start + t2 <= k:
            for i in range(start + t1, start + t2 - 1):
                for j in range(i + 1, start + t2):
                    Ediff = (1 / m) * np.sum(np.abs(x[:, i] - x[:, j])**p)
                    score += (1 / np.abs(i - j)) * (np.abs(y[i] - y[j])**p - Ediff)**2
                score_list.append(score)

    return score/(100_000), score_list

def calculate_crps(actuals, corrected_ensembles):
    """
    Calculate the Continuous Ranked Probability Score (CRPS).
    
    Parameters:
    actuals (numpy.ndarray): Actual observations
    corrected_ensembles (numpy.ndarray): Ensemble forecasts
    
    Returns:
    float: Mean CRPS score
    """
    try:
        crps = ps.crps_ensemble(actuals, corrected_ensembles)
        return np.mean(crps)
    except:
        crps = np.mean(ps.crps_ensemble(actuals, corrected_ensembles.T))
        return crps

def calculate_qss(actuals, taqr_results, quantiles):
    """
    Calculate the Quantile Skill Score (QSS) for multiple quantile forecasts.
    
    Parameters:
    actuals (numpy.ndarray): Actual observations
    taqr_results (numpy.ndarray): Predicted quantiles
    quantiles (list): List of quantile levels
    
    Returns:
    float: Mean QSS score
    """
    qss_scores = multi_quantile_skill_score(actuals, taqr_results, quantiles)
    return np.mean(qss_scores)

def multi_quantile_skill_score(y_true, y_pred, quantiles):
    """
    Calculate the Quantile Skill Score (QSS) for multiple quantile forecasts.

    Parameters:
    y_true (numpy.ndarray): True observed values
    y_pred (numpy.ndarray): Predicted quantile values
    quantiles (list): Quantile levels between 0 and 1

    Returns:
    numpy.ndarray: QSS for each quantile forecast
    """
    y_pred = np.array(y_pred)

    if y_pred.shape[0] > y_pred.shape[1]:
        y_pred = y_pred.T

    assert all(0 <= q <= 1 for q in quantiles), "All quantiles must be between 0 and 1"
    assert len(quantiles) == len(y_pred), "Number of quantiles must match inner dimension of y_pred"

    N = len(y_true)
    scores = np.zeros(len(quantiles))

    for i, q in enumerate(quantiles):
        E = y_true - y_pred[i]
        scores[i] = np.sum(np.where(E > 0, q * E, (1 - q) * -E))

    return scores / N

def run_r_script(X_filename, Y_filename, tau):
    """
    Run R script for quantile regression.
    
    Parameters:
    X_filename (str): Path to X data CSV file
    Y_filename (str): Path to Y data CSV file
    tau (float): Quantile level
    """
    import subprocess
    process = subprocess.Popen(["R", "--vanilla"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    
    r_script = f"""
    options(warn = -1)
    library(onlineforecast) 
    library(quantreg) 
    library(readr) 
    X_full <- read_csv("{X_filename}", col_names = FALSE, show_col_types = FALSE) 
    y <- read_csv("{Y_filename}", col_names = "y", show_col_types = FALSE) 
    X_full <- X_full[1:500,]
    data <- cbind(X_full, y[1:500,1]) 
    predictor_cols <- colnames(X_full) 
    formula_string <- paste("y ~ 0+", paste(predictor_cols, collapse = " + ")) 
    formula <- as.formula(formula_string) 
    rq_fit <- rq(formula, tau = {tau}, data = data ) 
    write.csv(rq_fit$coefficients, "rq_fit_coefficients.csv") 
    write.csv(rq_fit$residuals, "rq_fit_residuals.csv") 
    """
    
    for line in r_script.strip().split('\n'):
        process.stdin.write(line.encode('utf-8') + b"\n")

    process.stdin.close()
    process.stdout.read()
    process.terminate()

def remove_zero_columns(df):
    """Remove columns that contain only zeros from a DataFrame."""
    return df.loc[:, (df != 0).any(axis=0)]

def remove_zero_columns_numpy(arr):
    """Remove columns that contain only zeros or constant values from a numpy array."""
    return arr[:, (arr != 0).any(axis=0) & (arr != arr[0]).any(axis=0)]

def create_dataset_for_lstm(X, Y, time_steps):
    """
    Create a dataset suitable for LSTM training with multiple time steps.
    
    Parameters:
    X (numpy.ndarray): Input features
    Y (numpy.ndarray): Target values
    time_steps (list): List of time steps to include
    
    Returns:
    tuple: (X_lstm, Y_lstm) LSTM-ready datasets
    """
    X = np.array(X)
    Y = np.array(Y)

    Xs, Ys = [], []
    for i in range(len(X)):
        X_entry = []
        for ts in time_steps:
            if i - ts >= 0:
                X_entry.append(X[i - ts,:])
            else:
                X_entry.append(np.zeros_like(X[0,:]))
        Xs.append(np.array(X_entry))
        Ys.append(Y[i])
    return np.array(Xs), np.array(Ys)

class QuantileRegressionLSTM(tf.keras.Model):
    """
    LSTM-based model for quantile regression.
    
    Parameters:
    n_quantiles (int): Number of quantiles to predict
    units (int): Number of LSTM units
    n_timesteps (int): Number of time steps in input
    """
    
    def __init__(self, n_quantiles, units, n_timesteps, **kwargs):
        super().__init__(**kwargs)
        self.lstm = tf.keras.layers.LSTM(units, input_shape=(None, n_quantiles, n_timesteps), return_sequences=False)
        self.dense = tf.keras.layers.Dense(n_quantiles, activation='sigmoid')
        self.dense2 = tf.keras.layers.Dense(n_quantiles, activation='relu') 
        self.n_quantiles = n_quantiles
        self.n_timesteps = n_timesteps

    def call(self, inputs, training=None):
        x = self.lstm(inputs, training=training)
        x = self.dense(x)
        x = self.dense2(x)
        return x
    
    def get_config(self):
        config = super(QuantileRegressionLSTM, self).get_config()
        config.update({
            'n_quantiles': self.n_quantiles,
            'units': self.lstm.units,
            'n_timesteps': self.n_timesteps,
        })
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)

def quantile_loss_3(q, y_true, y_pred):
    """
    Calculate quantile loss for a single quantile.
    
    Parameters:
    q (float): Quantile level
    y_true (tensor): True values
    y_pred (tensor): Predicted values
    
    Returns:
    tensor: Quantile loss
    """
    y_true = tf.cast(y_true, tf.float32)
    y_pred = tf.cast(y_pred, tf.float32)
    y_true = tfp.stats.percentile(y_true, 100*q, axis=1)
    error = y_true - y_pred
    return tf.maximum(q * error, (q - 1) * error)

def quantile_loss_func(quantiles):
    """
    Create a loss function for multiple quantiles.
    
    Parameters:
    quantiles (list): List of quantile levels
    
    Returns:
    function: Loss function for multiple quantiles
    """
    def loss(y_true, y_pred):
        losses = []
        for i, q in enumerate(quantiles):
            loss = quantile_loss_3(q, y_true, y_pred[:,  i])
            losses.append(loss)
        return losses
    return loss

def map_range(values, input_start, input_end, output_start, output_end):
    """
    Map values from one range to another.
    
    Parameters:
    values (list): Values to map
    input_start (float): Start of input range
    input_end (float): End of input range
    output_start (float): Start of output range
    output_end (float): End of output range
    
    Returns:
    numpy.ndarray: Mapped values
    """
    mapped_values = []
    for value in values:
        proportion = (value - input_start) / (input_end - input_start)
        mapped_value = output_start + (proportion * (output_end - output_start))
        mapped_values.append(int(mapped_value))
    
    return np.array(mapped_values)

def legend_without_duplicate_labels(ax):
    """
    Create a legend without duplicate labels.
    
    Parameters:
    ax (matplotlib.axes.Axes): Axes object to create legend for
    """
    handles, labels = ax.get_legend_handles_labels()
    unique = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
    ax.legend(*zip(*unique))



import numpy as np

def remove_straight_line_outliers(ensembles):
    """
    Removes ensemble members that are perfectly straight lines (constant slope).
    
    Parameters:
        ensembles (numpy.ndarray): 2D array where rows are time steps and columns are ensemble members.
    
    Returns:
        numpy.ndarray: Filtered ensemble data without straight-line outliers.
    """
    # Calculate differences along the time axis
    differences = np.diff(ensembles, axis=0)
    
    # Identify columns where all differences are the same (perfectly straight lines)
    straight_line_mask = np.all(differences == differences[0, :], axis=0)
    
    # Remove the columns with perfectly straight lines
    return ensembles[:, ~straight_line_mask]

def train_model_lstm(quantiles, epochs: int, lr: float, batch_size: int, x, y, x_val, y_val, n_timesteps, data_name):
    """
    Train an LSTM model for quantile regression.

    Parameters:
    quantiles (list): List of quantile levels to predict
    epochs (int): Number of training epochs
    lr (float): Learning rate for optimizer
    batch_size (int): Batch size for training
    x (tensor): Training input data
    y (tensor): Training target data
    x_val (tensor): Validation input data
    y_val (tensor): Validation target data
    n_timesteps (int): Number of time steps in input sequence
    data_name (str): Name identifier for saving model artifacts

    Returns:
    tf.keras.Model: Trained LSTM model
    """
    model = QuantileRegressionLSTM(n_quantiles=len(quantiles), units=256, n_timesteps=n_timesteps)
    optimizer = tf.keras.optimizers.Adam(learning_rate=lr)

    @tf.function
    def train_step(x_batch, y_batch):
        with tf.GradientTape() as tape:
            y_pred = model(x_batch, training=True)
            losses = quantile_loss_func(quantiles)(y_batch, y_pred)
            total_loss = tf.reduce_mean(losses)
        
        grads = tape.gradient(total_loss, model.trainable_variables)
        optimizer.apply_gradients(zip(grads, model.trainable_variables))
        return total_loss
    
    @tf.function
    def val_step(x_batch, y_batch):
        y_pred = model(x_batch, training=False)
        losses = quantile_loss_func(quantiles)(y_batch, y_pred)
        total_loss = tf.reduce_mean(losses)
        return total_loss

    train_loss_history = []
    val_loss_history = []
    y_preds = []
    y_true = []

    for epoch in range(epochs):
        epoch_train_loss = 0.0
        epoch_val_loss = 0.0
        num_batches = 0
        
        # Training loop
        for i in range(0, len(x), batch_size):
            x_batch = x[i:i+batch_size]
            y_batch = y[i:i+batch_size]
            
            batch_train_loss = train_step(x_batch, y_batch)
            epoch_train_loss += batch_train_loss
            num_batches += 1
            
            y_preds.append(model(x_batch, training=False))
            y_true.append(y_batch)
        
        epoch_train_loss /= num_batches
        train_loss_history.append(epoch_train_loss)
        
        # Validation loop
        num_val_batches = 0
        for i in range(0, len(x_val), batch_size):
            x_val_batch = x_val[i:i+batch_size]
            y_val_batch = y_val[i:i+batch_size]
            
            batch_val_loss = val_step(x_val_batch, y_val_batch)
            epoch_val_loss += batch_val_loss
            num_val_batches += 1
        
        epoch_val_loss /= num_val_batches
        val_loss_history.append(epoch_val_loss)

        print(f"Epoch {epoch+1} Train Loss: {epoch_train_loss:.4f} Validation Loss: {epoch_val_loss:.4f}")

    y_preds_concat = tf.concat(y_preds, axis=0).numpy()
    y_true_concat = tf.concat(y_true, axis=0).numpy()

    return model

def one_step_quantile_prediction(X_input, Y_input, n_init, n_full, quantile=0.5, already_correct_size=False, n_in_X=5000):
    """
    Perform one-step quantile prediction using TAQR.
    
    This function takes the entire training set and, based on the last n_init observations,
    calculates residuals and coefficients for the quantile regression.

    Parameters:
    X_input (numpy.ndarray/pd.DataFrame): Input features
    Y_input (numpy.ndarray/pd.Series): Target values
    n_init (int): Number of initial observations for warm start
    n_full (int): Total number of observations to process
    quantile (float): Quantile level for prediction (default: 0.5)
    already_correct_size (bool): Whether input data is already correctly sized
    n_in_X (int): Number of observations to include in design matrix

    Returns:
    tuple: (predictions, actual values, coefficients)
    """
    assert n_init <= n_full - 2, "n_init must be less than n_full"

    if type(X_input) == pd.DataFrame:
        X_input = X_input.to_numpy()

    if type(Y_input) == pd.Series or type(Y_input) == pd.DataFrame:
        Y_input = Y_input.to_numpy()

    n, m = X_input.shape
    print("X_input shape: ", X_input.shape)
    
    full_length, p = X_input.shape

    X = X_input[:n_full, :].copy()
    Y = Y_input[:n_full]

    X_for_residuals = X[:n_init, :]
    Y_for_residuals = Y[:n_init]

    np.savetxt("X_for_residuals.csv", X_for_residuals, delimiter=",")
    np.savetxt("Y_for_residuals.csv", Y_for_residuals, delimiter=",")

    run_r_script("X_for_residuals.csv", "Y_for_residuals.csv", tau=quantile)

    def ignore_first_column(s):
        return float(s)

    residuals = np.genfromtxt(
        'rq_fit_residuals.csv', 
        delimiter=',', 
        skip_header=1, 
        usecols=(1,),
        converters={0: ignore_first_column}
    )

    beta_init = np.genfromtxt(
        'rq_fit_coefficients.csv', 
        delimiter=',', 
        skip_header=1, 
        usecols=(1,),
        converters={0: ignore_first_column}
    )

    print("len of beta_init: ", len(beta_init))
    print("There is: ", sum(residuals == 0), "zeros in residuals", "and", sum(abs(residuals) < 1e-8), "close to zeroes")
    print("p: ", p)

    beta_init = np.append(beta_init, np.ones(p-len(beta_init)))
    r_init = set_n_closest_to_zero(arr=residuals, n=len(beta_init))
    print(sum(r_init==0), "r_init zeros")

    X_full = np.column_stack((X, Y, np.random.choice([1,1], size=n_full)))
    IX = np.arange(p)
    Iy = p
    Iex = p + 1
    bins = np.array([-np.inf, np.inf])
    tau = quantile
    n_in_bin = int(1.0*full_length)
    print("n_in_bin", n_in_bin)

    n_input = n_in_X
    N, BETA, GAIN, Ld, Rny, Mx, Re, CON1, T = rq_simplex_final(X_full, IX, Iy, Iex, r_init, beta_init, n_input, tau, bins, n_in_bin)

    y_pred = np.sum((X_input[(n_input+2):(n_full), :] * BETA[1:,:]), axis=1)
    y_actual = Y_input[(n_input):(n_full-2)]
    print(y_pred.shape, "y_pred shape")
    print(y_actual.shape, "y_actual shape")

    y_actual_quantile = np.quantile(y_actual, quantile)
    return y_pred, y_actual, BETA

def run_taqr(corrected_ensembles, actuals, quantiles, n_init, n_full, n_in_X):
    """
    Run TAQR on corrected ensembles.

    Parameters:
    corrected_ensembles (numpy.ndarray): Shape (n_timesteps, n_ensembles)
    actuals (numpy.ndarray): Shape (n_timesteps,)
    quantiles (list): Quantiles to predict
    n_init (int): Number of initial timesteps for warm start
    n_full (int): Total number of timesteps
    n_in_X (int): Number of timesteps in design matrix

    Returns:
    list: TAQR results for each quantile
    """
    if type(actuals) == pd.Series or type(actuals) == pd.DataFrame:
        # remove nans from actuals
        actuals = actuals.to_numpy()
        actuals[np.isnan(actuals)] = 0
    else:
        actuals[np.isnan(actuals)] = 0

    taqr_results = []
    for q in quantiles:
        print("running TAQR for quantile: ", q)
        y_pred, _, _ = one_step_quantile_prediction(corrected_ensembles, actuals, n_init=n_init, n_full=n_full, quantile=q, already_correct_size=True, n_in_X=n_in_X)
        taqr_results.append(y_pred)

    return taqr_results

def pipeline(X, y, name="TEST", training_size=0.8, epochs=100, timesteps_for_lstm=[0,1,2,6,12,24,48], **kwargs):
    """
    Main pipeline for NABQR model training and evaluation.

    The pipeline:
    1. Trains an LSTM network to correct the provided ensembles
    2. Runs TAQR algorithm on corrected ensembles to predict observations
    3. Saves results and model artifacts

    Parameters:
    X (pd.DataFrame/numpy.ndarray): Shape (n_samples, n_features) - Ensemble data
    y (pd.Series/numpy.ndarray): Shape (n_samples,) - Observations
    name (str): Dataset identifier
    training_size (float): Proportion of data for training
    epochs (int): Number of LSTM training epochs
    timesteps_for_lstm (list): Time steps for LSTM input
    **kwargs: Additional parameters (e.g., quantiles_taqr)

    Returns:
    None: Results are saved to files
    """
    # Data preparation
    actuals = y
    ensembles = X
    X_y = np.concatenate((X, y.reshape(-1,1)), axis=1)
    
    if isinstance(y, pd.Series):
        idx = y.index
    elif isinstance(X, pd.DataFrame):
        idx = X.index
    else:
        idx = pd.RangeIndex(start=0, stop=len(y), step=1)

    train_size = int(training_size * len(actuals))
    ensembles = pd.DataFrame(ensembles, index=idx)
    ensembles.index = pd.to_datetime(ensembles.index, utc=False).tz_localize(None)
    actuals = pd.DataFrame(actuals, index=idx)
    actuals.index = pd.to_datetime(actuals.index, utc=False).tz_localize(None)
    common_index = ensembles.index.intersection(actuals.index)
    X_y = pd.DataFrame(X_y, index=idx)
    X_y.index = pd.to_datetime(X_y.index, utc=False).tz_localize(None)
    ensembles = ensembles.loc[common_index]
    actuals = actuals.loc[common_index]
    X_y = X_y.loc[common_index]

    print(ensembles)
    timesteps = timesteps_for_lstm
    Xs, X_Ys = create_dataset_for_lstm(ensembles, X_y, timesteps_for_lstm)

    # Handle NaN values
    if np.isnan(Xs).any():
        print("Xs has NaNs")
        Xs[np.isnan(Xs).any(axis=(1,2))] = 0
    if np.isnan(X_Ys).any():
        print("X_Ys has NaNs")
        X_Ys[np.isnan(X_Ys).any(axis=1)] = 0

    # Data standardization
    XY_s_max_train = np.max(X_Ys[:train_size])
    XY_s_min_train = np.min(X_Ys[:train_size])

    X_Ys_scaled_train = (X_Ys[:train_size] - XY_s_min_train) / (XY_s_max_train - XY_s_min_train)
    Xs_scaled_train = (Xs[:train_size] - XY_s_min_train) / (XY_s_max_train - XY_s_min_train)

    validation_size = 100
    X_Ys_scaled_validation = (X_Ys[train_size:(train_size+validation_size)] - XY_s_min_train) / (XY_s_max_train - XY_s_min_train)
    Xs_scaled_validation = (Xs[train_size:(train_size+validation_size)] - XY_s_min_train) / (XY_s_max_train - XY_s_min_train)

    # Train LSTM model
    quantiles_lstm = np.linspace(0.05, 0.95, 20)
    model = train_model_lstm(quantiles=quantiles_lstm, epochs=epochs, 
                           lr=1e-3, batch_size=50, 
                           x=tf.convert_to_tensor(Xs_scaled_train), 
                           y=tf.convert_to_tensor(X_Ys_scaled_train),
                           x_val=tf.convert_to_tensor(Xs_scaled_validation),
                           y_val=tf.convert_to_tensor(X_Ys_scaled_validation),
                           n_timesteps=timesteps,
                           data_name=f"{name}_LSTM_epochs_{epochs}")
    
    # Save model
    try:    
        today = dt.datetime.today().strftime('%Y-%m-%d')
        model.save(f"Model_{name}_{epochs}_{today}.keras")
    except:
        model.save(f"Models_{name}_{epochs}.keras")

    # Generate predictions
    Xs_scaled_test = (Xs[train_size:] - XY_s_min_train) / (XY_s_max_train - XY_s_min_train)
    corrected_ensembles = model(Xs_scaled_test)
    corrected_ensembles = corrected_ensembles * (XY_s_max_train - XY_s_min_train) + XY_s_min_train
    actuals_out_of_sample = actuals[train_size:]
    test_idx = idx[train_size:]

    # Run TAQR
    quantiles_taqr = kwargs.get("quantiles_taqr", [0.1, 0.3, 0.5, 0.7, 0.9])
    n_full = len(actuals_out_of_sample)
    n_init = int(0.25*n_full)
    print("n_init, n_full: ", n_init, n_full)

    corrected_ensembles = corrected_ensembles.numpy()
    corrected_ensembles = remove_zero_columns_numpy(corrected_ensembles)
    corrected_ensembles = remove_straight_line_outliers(corrected_ensembles)

    n_in_X = n_init
    taqr_results = run_taqr(corrected_ensembles, actuals_out_of_sample, quantiles_taqr, n_init, n_full, n_in_X)
    actuals_out_of_sample = actuals_out_of_sample[(n_init+1):(n_full-1)]

    corrected_ensembles = corrected_ensembles[(n_init+1):(n_full-1)]
    idx_to_save = test_idx[(n_init+1):(n_full-1)]

    # Save results
    data_source = f"{name}"
    today = dt.datetime.today().strftime('%Y-%m-%d')

    np.save(f"results_{today}_{data_source}_actuals_out_of_sample.npy", actuals_out_of_sample)
    df_corrected_ensembles = pd.DataFrame(corrected_ensembles, index=idx_to_save)
    df_corrected_ensembles.to_csv(f"results_{today}_{data_source}_corrected_ensembles.csv")
    np.save(f"results_{today}_{data_source}_taqr_results.npy", taqr_results)



import numpy as np
import pandas as pd

def set_n_smallest_to_zero(arr, n):
    """
    Sets the n smallest elements in an array to zero.
    
    Parameters:
    arr (array-like): Input array of numbers
    n (int): Number of smallest elements to set to zero
    
    Returns:
    numpy.ndarray: Modified array with n smallest elements set to zero
    """
    if n <= 0:
        return arr
    
    if n >= len(arr):
        return [0] * len(arr)
    
    # Find the nth smallest element
    nth_smallest = sorted(arr)[n-1]
    
    # Set elements smaller than or equal to nth_smallest to zero
    modified_arr = [0 if x <= nth_smallest else x for x in arr]
    modified_arr = np.array(modified_arr)
    return modified_arr

def set_n_closest_to_zero(arr, n):
    """
    Sets the n elements closest to zero in an array to zero.
    
    Parameters:
    arr (array-like): Input array of numbers
    n (int): Number of elements closest to zero to set to zero
    
    Returns:
    numpy.ndarray: Modified array with n elements closest to zero set to zero
    """
    if n <= 0:
        return arr
    
    if n >= len(arr):
        return [0] * len(arr)
    
    # Find the absolute values of the elements
    abs_arr = np.abs(arr)
    
    # Find the indices of the n elements closest to zero
    closest_indices = np.argpartition(abs_arr, n)[:n]
    
    # Set the elements closest to zero to zero
    modified_arr = arr.copy()
    modified_arr[closest_indices] = 0
    
    return modified_arr

def quantile_score(p, z, q):
    """
    Calculate the Quantile Score (QS) for a given probability and set of observations and quantiles.
    Implementation based on Fauer et al. (2021): "Flexible and consistent quantile estimation for
    intensity–duration–frequency curves"

    Parameters:
    p (float): The probability level (between 0 and 1)
    z (numpy.ndarray): The observed values
    q (numpy.ndarray): The predicted quantiles

    Returns:
    float: The Quantile Score (QS)
    """
    u = z - q
    rho = np.where(u > 0, p * u, (p - 1) * u)
    return np.sum(rho)  

def simulate_correlated_ar1_process(n, phi, sigma, m, corr_matrix=None, offset=None, smooth="no"):
    """
    Simulate a correlated AR(1) process with multiple dimensions.
    
    Parameters:
    n (int): Number of time steps to simulate
    phi (float): AR(1) coefficient (persistence parameter)
    sigma (float): Standard deviation of the noise
    m (int): Number of dimensions/variables
    corr_matrix (numpy.ndarray, optional): Correlation matrix between dimensions. Defaults to identity matrix
    offset (numpy.ndarray, optional): Offset vector for each dimension. Defaults to zero vector
    smooth (int or str, optional): Number of initial time steps to discard for smoothing. Defaults to "no"
    
    Returns:
    tuple: (simulated_ensembles, actuals) where simulated_ensembles is the AR(1) process 
           and actuals is the median of ensembles with added noise
    """
    if offset is None:
        offset = np.zeros(m)
    elif len(offset) != m:
        raise ValueError("Length of offset array must be equal to m")
    
    if corr_matrix is None:
        corr_matrix = np.eye(m)  # Default to no correlation (identity matrix)
    elif corr_matrix.shape != (m, m):
        raise ValueError("Correlation matrix must be of shape (m, m)")

    # Ensure the covariance matrix is positive semi-definite
    cov_matrix = sigma**2 * corr_matrix
    L = np.linalg.cholesky(cov_matrix)  # Cholesky decomposition

    if isinstance(smooth, int):
        ensembles = np.zeros((n + smooth, m))
        ensembles[0] = np.random.multivariate_normal(np.zeros(m), cov_matrix)

        for t in range(1, n + smooth):
            noise = np.random.multivariate_normal(np.zeros(m), cov_matrix)
            ensembles[t] = phi * ensembles[t-1] + noise

        # Extract the smoothed part of the ensembles
        smoothed_ensembles = ensembles[smooth:]

        return smoothed_ensembles + offset, np.median(smoothed_ensembles + offset, axis=1) + np.random.normal(0, sigma/2, n)

    else:
        ensembles = np.zeros((n, m))
        ensembles[0] = np.random.multivariate_normal(np.zeros(m), cov_matrix)

        for t in range(1, n):
            noise = np.random.multivariate_normal(np.zeros(m), cov_matrix)
            ensembles[t] = phi * ensembles[t-1] + noise
        return ensembles + offset, np.median(ensembles+ offset, axis=1) + np.random.normal(0, sigma/2, n)

 


"""
Time-Adaptive Quantile Regression (TAQR) Implementation

This module provides the core implementation of the TAQR algorithm based on
Møller's thesis "Modeling of Uncertainty in Wind Energy Forecast" (2006).
It implements an adaptive simplex algorithm for quantile regression problems.
"""

import numpy as np
import scipy.linalg
import time
from scipy import linalg as la
from sklearn.linear_model import QuantileRegressor
from helper_functions import *
import pandas as pd

def opdatering_final(X, Xny, IX, Iy, Iex, Ih, Ihc, beta, Rny, K, n, xB, P, tau, i, bins, n_in_bin):
    """
    Updates the design matrix and related parameters for the adaptive quantile regression.
    
    Parameters:
    X (numpy.ndarray): Full data matrix
    Xny (numpy.ndarray): Current design matrix
    IX (numpy.ndarray): Index set for design matrix columns
    Iy (int): Index for response variable
    Iex (int): Index for grouping variable
    Ih (numpy.ndarray): Index set for basic variables
    Ihc (numpy.ndarray): Index set for non-basic variables
    beta (numpy.ndarray): Current coefficient estimates
    Rny (numpy.ndarray): Residuals array
    K (int): Number of explanatory variables
    n (int): Current number of observations
    xB (numpy.ndarray): Basic solution
    P (numpy.ndarray): Sign vector
    tau (float): Quantile level
    i (int): Current iteration
    bins (numpy.ndarray): Bin boundaries
    n_in_bin (int): Maximum number of observations per bin
    
    Returns:
    tuple: Updated parameters (Ih, Ihc, xB, Xny, Rny, P, n, i)
    """
    Xny[Xny == np.inf] = 1
    Xny[Xny == -np.inf] = 1

    Xny = np.vstack((Xny, X[n + i - 2, :]))
    Index = np.arange(Xny[:-1,Iy].shape[0])
    j = 1
    
    while Xny[-1, Iex] > bins[j]:
        j += 1
    j -= 1
    
    In = Index[(Xny[:-1, Iex] > bins[j]) & (Xny[:-1, Iex] <= bins[j + 1])]
    
    if In.size > 0:
        Leav = np.min(In)
    else:
        Leav = 1

    minIL, Inmin = np.min(np.abs(Ih - Leav)), np.argmin(np.abs(Ih - Leav))

    if minIL == 0 and len(In) == n_in_bin:
        cB = (P < 0) + P * tau
        invXh = np.linalg.inv(Xny[Ih, IX])
        g = -(P * (Xny[Ihc, IX] @ invXh[:, Inmin])).T @ cB
        h = np.vstack((invXh[:, Inmin], -P * (Xny[Ihc, IX] @ invXh[:, Inmin]))) @ np.sign(g)
        sigma = np.zeros(n - K)
        hm = h[K:]
        xBm = xB[K:]
        xBm[xBm < 0] = 0
        tolerance = 1e-10
        sigma[hm > tolerance] = xBm[hm > tolerance] / hm[hm > tolerance]
        sigma[hm <= tolerance] = np.inf
        alpha, q = np.min(sigma), np.argmin(sigma)

        z = xB - alpha * h
        Ihm = Ih[Inmin]
        Ih[Inmin] = Ihc[q]
        Ihc[q] = Ihm

        xB = z
        xB[q + K] = alpha

        P[q] = np.sign(g) + (g == 0)
        Ih = np.sort(Ih)
        Ihc, IndexIhc = np.sort(Ihc), np.argsort(Ihc)
        P = P[IndexIhc]
        xBm = xB[K:]
        xBm = xBm[IndexIhc]
        xB[K:] = xBm
        beta = xB[:K]

    rny = X[n + i - 2, Iy] - X[n + i - 2, IX] @ beta
    Rny = np.append(Rny, rny)

    if rny < 0:
        P = np.append(P, -1)
        xB = np.append(xB, -rny)
    else:
        P = np.append(P, 1)
        xB = np.append(xB, rny)
    Ihc = np.append(Ihc, n)

    if len(In) == n_in_bin:
        i += 1
        Stay = np.ones(len(Ihc), dtype=bool)
        Stay[Ihc == Leav] = False
        P = P[Stay]
        Ihc = Ihc[Stay]
        Xny = Xny[np.sort(np.hstack((Ih, Ihc))), :]
        xBm = xB[K:]
        xBm = xBm[Stay]
        xB = xB[:-1]
        xB[K:] = xBm
        Ihc[Ihc > Leav] -= 1
        Ih[Ih > Leav] -= 1
    else:
        n += 1

    return Ih, Ihc, xB, Xny, Rny, P, n, i

def rq_initialiser_final(X, r, beta, n):
    """
    This function initializes the parameters needed for the simplex algorithm based on initial solution.
    If the number of zero elements in r is equal to rank(X), then the work is essentially done.
    Otherwise, the if statement will find the index set Ih s.t. X(Ih)*beta=y(Ih) and X(Ih)^(-1)*y(Ih)=beta,
    the important note being that X(Ih) has an inverse.
    This is done by using the LU transform of a non-quadratic matrix.
    
    Parameters:
    X (numpy.ndarray): Design matrix
    r (numpy.ndarray): Initial residuals
    beta (numpy.ndarray): Initial coefficients
    n (int): Number of observations
    
    Returns:
    tuple: (xB, Ih, Ihc, P) Basic solution, basic indices, non-basic indices, and sign vector
    """
    Index = np.arange(n)
    if np.sum(r == 0) > len(beta):
        Lr0 = np.sum(r == 0)
        rs, Irs = np.sort(np.abs(r)), np.argsort(np.abs(r))
        Irs = Irs[:Lr0]
        Xh = X[Irs, :]
        P, L, U = scipy.linalg.lu(Xh, permute_l=False)
        In = np.arange(Lr0)
        rI = np.zeros(Lr0 - len(beta), dtype=int)
        for i in range(len(beta), Lr0):
            rI[i - len(beta)] = In[P[:, i] == 1]
        r[Irs[rI]] = 1e-15

    index_to_index = r.flatten() == 0
    if len(index_to_index) < len(Index):
        index_to_index = np.append(index_to_index, np.zeros(len(Index) - len(index_to_index), dtype=bool))

    Ih = Index[index_to_index]
    Ihc = np.setdiff1d(Index[:len(r)], Ih)
    P = np.sign(r[Ihc])
    r[np.abs(r) < 1e-15] = 0
    xB = np.hstack((beta, np.abs(r[Ihc])))

    return xB, Ih, Ihc, P

def rq_simplex_alg_final(Ih, Ihc, n, K, xB, Xny, IH, P, tau):
    """
    Performs one step of the simplex algorithm for quantile regression.
    
    Parameters:
    Ih (numpy.ndarray): Index set for basic variables
    Ihc (numpy.ndarray): Index set for non-basic variables
    n (int): Number of observations
    K (int): Number of explanatory variables
    xB (numpy.ndarray): Basic solution
    Xny (numpy.ndarray): Design matrix
    IH (numpy.ndarray): History of index sets
    P (numpy.ndarray): Sign vector
    tau (float): Quantile level
    
    Returns:
    tuple: Algorithm parameters for the next iteration
    """
    invXh = la.inv(Xny[Ih, :])
    cB = (P < 0) + P * tau
    cC = np.vstack((np.ones(K) * tau, np.ones(K) * (1 - tau))).reshape(-1,1)
    IB2 = -np.dot(P.reshape(-1, 1) * (np.ones((1, K)) * Xny[Ihc, :]), invXh)
    g = IB2.T @ cB
    d = cC - np.vstack((g, -g)).reshape(-1,1)
    d[np.abs(d) < 1e-15] = 0
    d = d.flatten()

    md, s = np.sort(d), np.argsort(d)
    s = s[md < 0]
    md = md[md < 0]
    c = np.ones(len(s))
    c[s > (K-1)] = -1
    C = np.diag(c)
    s[s > (K-1)] -= K
    h = np.vstack([invXh[:,s], IB2[:,s]]) @ C
    alpha = np.zeros(len(s))
    q = np.zeros(len(s), dtype=int)
    xm = xB[K:]
    xm[xm < 0] = 0
    hm = h[K:,:]
    cq = np.zeros(len(s))
    tol1 = 1e-12

    for k in range(len(s)):
        sigma = xm.copy()
        sigma[hm[:, k] > tol1] = xm[hm[:, k] > tol1] / hm[hm[:, k] > tol1, k]
        sigma[hm[:, k] <= tol1] = np.inf
        alpha[k], q[k] = np.min(sigma), np.argmin(sigma)
        cq[k] = c[k]

    gain = md * alpha
    Mgain, IMgain = np.sort(gain), np.argsort(gain)
    CON = np.inf
    j = 0

    if len(gain) == 0:
        gain = 1
    else:
        while CON > 1e6 and j < len(s):
            j += 1
            IhMid = Ih.copy()
            shifter = 0
            IhMid[s[IMgain[j - 1 + shifter]]] = Ihc[q[IMgain[j - 1 + shifter]]]
            IhMid = np.sort(IhMid)

            if IH.shape[1] <= 1:
                IH = IH.T.reshape(-1,1)
            IhMid = IhMid.reshape(1,-1)

            if np.min(np.sum(np.abs(IH-IhMid.T * np.ones((1,IH.shape[1]))), axis=0)) == 0:
                CON = np.inf
            else:
                CON = np.linalg.cond(Xny[(IhMid.flatten()), :])

        s = s[IMgain[j - 1 + shifter]]
        q = q[IMgain[j - 1 + shifter]]
        cq = cq[IMgain[j - 1 + shifter]]
        alpha = alpha[IMgain[j - 1 + shifter]]
        IH = np.hstack((IH, IhMid.T))
        h = h[:, IMgain[j - 1 + shifter]]
        gain = gain[IMgain[j - 1 + shifter]]
        md = md[IMgain[j - 1 + shifter]]

    return CON, s, q, gain, md, alpha, h, IH, cq

def rq_purify_final(xB, Ih, Ihc, P, K, Xny, yny):
    """
    This function takes care of infeasible points in a simplex formulation
    of a quantile regression problem. The underlying assumption in this
    function is that there are no restrictions in the problem.

    The updating can therefore be done by recalculating all residuals and
    coefficients.

    The assumption is further that we are in a position s.t.
    Xny*Xny(Ih)^(-1)*yny(Ih)=yny+residuals
    K=rank(Xny)

    Parameters:
    xB (numpy.ndarray): Basic solution
    Ih (numpy.ndarray): Index set for basic variables
    Ihc (numpy.ndarray): Index set for non-basic variables
    P (numpy.ndarray): Sign vector
    K (int): Number of explanatory variables
    Xny (numpy.ndarray): Design matrix
    yny (numpy.ndarray): Response vector
    
    Returns:
    tuple: (xB, P) Updated basic solution and sign vector
    """
    invXh = np.linalg.inv(Xny[Ih, :])
    xB = np.hstack((invXh @ yny[Ih], yny[Ihc] - Xny[Ihc, :] @ invXh @ yny[Ih]))
    P = np.sign(xB[K:])
    P[P == 0] = 1
    xB[K:] = np.abs(xB[K:])
    return xB, P

def rq_simplex_final(X, IX, Iy, Iex, r, beta, n, tau, bins, n_in_bin):
    """
    rq_simplex_final calculates the solution to an adaptive simplex
    algorithm for a quantile regression problem. The function uses
    knowledge of the solution at time t to calculate the solution at
    time t+1.

    The basic idea is that the solution to the quantile regression
    problem can be written as:
    y(t) = X(t)'*beta + r(t)

    where beta = X(h)^(-1)*y(h) for some index set h. Simplex algorithm
    is now used to calculate the optimal h at time t+1 based on the solution
    at time t. So basically, the function uses h(t) as a starting guess for
    the simplex algorithm to iterate to h(t+1). This function is described
    in the master thesis "Modelling of Uncertainty in Wind Energy Forecast"
    by Jan K. Moeller, URL http://www.imm.dtu.dk/pubdb/p.php?4428.

    Inputs to the function:
    X        : The design matrix for the linear quantile regression problem,
               or rather it contains the design matrix, see the three next
               input variables.
    IX       : An index set, it refers to the columns of X which is the
               design matrix.
    Iy       : One index refers to the matrix X, the column Iy of X contains
               the response corresponding to the explanatory variables in the
               design matrix.
    Iex      : An index referring to a grouping variable, refers to a column of
               X this may or may not be a part of the design matrix. This is
               used in the updating algorithm.
    r        : The residuals from a solution to the quantile regression
               problem based on the first n (see below) elements of X. Such a
               solution could be obtained by the "rq" method in "R". This is
               only used to initialize the solution.
    beta     : The solution corresponding to the residuals in the vector r
               above.
    n        : The number of elements in r, i.e., y(1:n) = X(1:n,IX)*beta + r(1:n).
    tau      : The required probability, the solution in r and beta should be
               based on this probability.
    bins     : A vector defining a partition of an interval that covers all
               elements in X(1:end,Iex). This is used for the updating
               procedure of the design matrix. If bins = [-Inf,Inf], then the
               updating procedure is on a gliding window.
    n_in_bin : Number of elements in the bins defined above, this is one
               number, so the number of elements will be the same in each of
               the bins after some time.

    Remarks:
    This implementation requires that the number of elements in each bin of
    the initial solution is less than or equal to n_in_bin.

    Output:
    N    : The number of simplex steps to get the solutions at time t+1
           given the solution at time t. This is a vector.
    BETA : A matrix with each column being the solution to the quantile
           regression problem, corresponding to the matrix X(n+1:end,1:end).
    GAIN : This is a parameter used only to analyze the method. It is the
           gain in the loss function in each simplex step. If this is very
           large at some points, it can be taken as a sign of the problem
           being ill-posed.
    Ld   : This is only used for analyzing the algorithm in its own right. It
           gives the number of descent directions in each simplex step.
    Rny  : This is the residual of the one-step-ahead prediction of the
           method.
    Mx   : This is only used for analyzing the algorithm in its own right. It
           is the minimum of the constraint solution to the simplex
           formulation. This should be larger than or equal to zero. If this
           becomes large in absolute value, then it is an indication of the
           algorithm having problems at this point.
    Re   : The reliability on the training set for each point along the
           solution. If everything is good, then this should follow Theorem
           2.3 in the reference mentioned above. I.e., this is close to tau at
           all points.
    CON1 : The condition number of the matrices X(h(t)).
    T    : The time used for each iteration.

    References:
    [1] J. K. Møller (2006), Modeling of Uncertainty in Wind Energy
           Forecast. Master Thesis, Informatics and Mathematical Modelling,
           Technical University of Denmark. Available at
           http://www.imm.dtu.dk/pubdb/p.php?4428.

    [2] H. B. Nielsen (1999), Algorithms for Linear Optimization, an
           Introduction. Course note for the DTU course "Optimization and Data
           Fitting 2". Available at http://www.imm.dtu.dk/courses/02611/
    """
    GAIN = np.array([0])
    Rny = np.array([0])
    Ld = np.array([0])
    mx = 0
    T = np.zeros(len(X[:, Iy]))
    Mx = np.zeros(len(X[:, Iy]))
    N_num_of_simplex_steps = []
    N_size_of_the_design_matrix_at_time_k = []
    CON1 = np.array([0])
    BETA = beta.reshape(1, -1)

    K = len(beta)
    Xny = X[:(n), :]
    H = np.zeros((1, K))
    tolmx = 1e-15
    j = 0
    LX = len(X[:, Iy])
    i = 2
    k = 0

    xB, Ih, Ihc, P = rq_initialiser_final(Xny[:, IX], r, beta, n)
    Re = np.zeros(LX - n)

    while i + n < LX:
        k += 1
        t = time.time()
        
        Re[k] = np.sum(P < 0) / n
        mx = np.min(xB[K:])

        if j > 0 and mx < -tolmx:
            xB, P = rq_purify_final(xB, Ih, Ihc, P, K, Xny[:, IX], Xny[:, Iy])
            mx = np.min(xB[K:])

        Mx[k] = mx
        j = 0
        beta = xB[:K]
        BETA = np.vstack((BETA, beta))

        Ih, Ihc, xB, Xny, Rny, P, n, i = opdatering_final(
            X, Xny, IX, Iy, Iex, Ih, Ihc, beta, Rny, K, n, xB, P, tau, i, bins, n_in_bin
        )

        IH = Ih.reshape(-1, 1)
        CON, s, q, gain, md, alpha, h, IH, cq = rq_simplex_alg_final(
            Ih, Ihc, n, K, xB, Xny[:, IX], IH, P, tau
        )
        CON1 = np.append(CON1, CON)

        while gain <= 0 and md < 0 and j < 24 and CON < 1e6:
            GAIN = np.append(GAIN, gain)
            j += 1
      
            z = xB - alpha * h
            
            IhM = Ih[s]
            IhcM = Ihc[q]
            Ih[s] = IhcM
            Ihc[q] = IhM
            P[q] = cq

            xB = z
            xB[q + K] = alpha

            Ih = np.sort(Ih)
            Ihc, IndexIhc = np.sort(Ihc), np.argsort(Ihc)
            P = P[IndexIhc]
            xBm = xB[K:]
            xBm = xBm[IndexIhc]
            xB[K:] = xBm

            CON, s, q, gain, md, alpha, h, IH, cq = rq_simplex_alg_final(
                Ih, Ihc, n, K, xB, Xny[:, IX], IH, P, tau
            )
            CON1 = np.append(CON1, CON)

        N_num_of_simplex_steps.append(j)
        N_size_of_the_design_matrix_at_time_k.append(n)
        T[k] = time.time() - t
    
    N = np.hstack([N_num_of_simplex_steps, N_size_of_the_design_matrix_at_time_k])
    return N, BETA, GAIN[1:], Ld, Rny, Mx, Re, CON1[1:], T

def one_step_quantile_prediction(X_input, Y_input, n_init, n_full, quantile=0.5, already_correct_size=False, n_in_X=5000):
    """
    As input, this function should take the entire training set, and based on the last n_init observations,
    calculate residuals and coefficients for the quantile regression.
    
    Parameters:
    X_input (numpy.ndarray): Input features matrix
    Y_input (numpy.ndarray): Target values array
    n_init (int): Number of initial observations for training
    n_full (int): Total number of observations to use
    quantile (float): Quantile level to predict
    already_correct_size (bool): Whether inputs are already correctly sized
    n_in_X (int): Number of observations to use in X
    
    Returns:
    tuple: (y_pred, y_actual, BETA) Predictions, actual values, and coefficients
    """
    assert n_init <= n_full - 2, "n_init must be less than n_full"

    if type(X_input) == pd.DataFrame:
        X_input = X_input.to_numpy()

    if type(Y_input) == pd.Series or type(Y_input) == pd.DataFrame:
        Y_input = Y_input.to_numpy()

    n, m = X_input.shape
    full_length, p = X_input.shape

    X = X_input[:n_full, :].copy()
    Y = Y_input[:n_full]

    X_for_residuals = X[:n_init, :]
    Y_for_residuals = Y[:n_init]

    np.savetxt("X_for_residuals.csv", X_for_residuals, delimiter=",")
    np.savetxt("Y_for_residuals.csv", Y_for_residuals, delimiter=",")

    run_r_script("X_for_residuals.csv", "Y_for_residuals.csv", tau=quantile)

    def ignore_first_column(s):
        return float(s)

    residuals = np.genfromtxt(
        'rq_fit_residuals.csv', 
        delimiter=',', 
        skip_header=1, 
        usecols=(1,),
        converters={0: ignore_first_column}
    )

    beta_init = np.genfromtxt(
        'rq_fit_coefficients.csv', 
        delimiter=',', 
        skip_header=1, 
        usecols=(1,),
        converters={0: ignore_first_column}
    )

    beta_init = np.append(beta_init, np.ones(p-len(beta_init)))
    r_init = set_n_closest_to_zero(arr=residuals, n=len(beta_init))

    X_full = np.column_stack((X, Y, np.random.choice([1,1], size=n_full)))
    IX = np.arange(p)
    Iy = p
    Iex = p + 1
    bins = np.array([-np.inf, np.inf])
    tau = quantile
    n_in_bin = int(1.0*full_length)

    n_input = n_in_X
    N, BETA, GAIN, Ld, Rny, Mx, Re, CON1, T = rq_simplex_final(X_full, IX, Iy, Iex, r_init, beta_init, n_input, tau, bins, n_in_bin)

    y_pred = np.sum((X_input[(n_input+2):(n_full), :] * BETA[1:,:]), axis=1)
    y_actual = Y_input[(n_input):(n_full-2)]

    return y_pred, y_actual, BETA

